
#include "libvdev/misc.h"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <poll.h>
#include <sys/signalfd.h>
#include <stdbool.h>
#include <sys/inotify.h>
#include <dirent.h>

const char *progname = "vdev_inotify_wait";

/* Structure to keep track of monitored directories */
typedef struct {
  
   /* Path of the directory */
   char *path;
  
   /* inotify watch descriptor */
   int wd;
  
} monitored_t;

/* Size of buffer to use when reading inotify events */
#define INOTIFY_BUFFER_SIZE 8192

/* Enumerate list of FDs to poll */
enum {
  
   FD_POLL_SIGNAL = 0,
   FD_POLL_INOTIFY,
   FD_POLL_MAX
};

/* Array of directories being monitored */
static monitored_t *monitors;
static int n_monitors;

static
void shutdown_inotify (int inotify_fd)
{
   for (unsigned i = 0; i < n_monitors; ++i) {
      
      free (monitors[i].path);
      inotify_rm_watch (inotify_fd, monitors[i].wd);
   }
   
   free (monitors);
}

static
int initialize_inotify (int argc, char **argv)
{
   int inotify_fd;

   /* Create new inotify device */
   if ((inotify_fd = inotify_init ()) < 0) {
      
      vdev_error("%s: Couldn't setup new inotify device: '%s'\n", progname, strerror (errno));
      return -1;
   }

   /* Allocate array of monitor setups */
   n_monitors = argc - 1;
   monitors = malloc (n_monitors * sizeof (monitored_t));

   /* Loop all input directories, setting up watches */
   for (unsigned i = 0; i < n_monitors; ++i) {
      
      monitors[i].path = vdev_strdup_or_null (argv[i + 1]);
      if ((monitors[i].wd = inotify_add_watch (inotify_fd, monitors[i].path, IN_CREATE)) < 0) {
        
          vdev_error ("%s: Couldn't add monitor in directory '%s': '%s'\n", progname, monitors[i].path, strerror (errno));
          exit (EXIT_FAILURE);
      }
      
      //printf ("Started monitoring directory '%s'...\n", monitors[i].path);
   }

   return inotify_fd;
}

static
int initialize_signals (void)
{
   int signal_fd;
   sigset_t sigmask;

   /* We want to handle SIGINT and SIGTERM in the signal_fd, so we block them. */
   sigemptyset (&sigmask);
   sigaddset (&sigmask, SIGINT);
   sigaddset (&sigmask, SIGTERM);

   if (sigprocmask (SIG_BLOCK, &sigmask, NULL) < 0) {
      
      vdev_error ("%s: Couldn't block signals: '%s'\n", progname, strerror (errno));
      return -1;
   }

   /* Get new FD to read signals from it */
   if ((signal_fd = signalfd (-1, &sigmask, 0)) < 0) {
      
      vdev_error ("%s: Couldn't setup signal FD: '%s'\n", progname, strerror (errno));
      return -1;
   }

   return signal_fd;
}


int main (int argc, char **argv)
{	
   int rc = 0;	
   DIR *dirp;
   struct dirent *E;
   const char *path;
   struct pollfd pfd[FD_POLL_MAX];
   bool is_userspace = false;
   
   int signal_fd;
   int inotify_fd;
   
   /* Input arguments... */
   if (argc < 2) {
      
      fprintf (stderr, "Usage: %s mode(0|1) directory1 [directory2 ...]\n", argv[0]);
      exit (EXIT_FAILURE);
   }
   
   path = vdev_strdup_or_null(argv[1]);
 
   dirp = opendir(path);
   if(!dirp) {
      fprintf(stderr, "%s error opening current directory: %s\n", progname, strerror(errno));
      exit(EXIT_FAILURE);
   }

   while( (errno=0, E=readdir(dirp)) ) {
	   
      struct stat devstat;
      if( lstat(E->d_name, &devstat) ) {
	     fprintf(stderr, "%s warning: create unknown device %s\n", progname, E->d_name);
	     closedir(dirp);
	     exit(EXIT_FAILURE);
	  }
      
      if( (devstat.st_mode & S_IFMT) == S_IFDIR ) {
		
	     if(!strcmp(E->d_name, ".") || !strcmp(E->d_name, ".."))
		    continue;
		
         char str[4097];            ///
         struct stat devstat;
         strncpy_t(str, sizeof(str), path, strlen(path));
         strcat(str, "/");
         strcat(str, E->d_name);
          
         if (lstat(str, &devstat) == 0) {
              
            //printf("created %s Ownership: UID=%ld   GID=%ld\n", inot_evt->name, (long) devstat.st_uid, (long) devstat.st_gid);
            if((uintmax_t)devstat.st_uid != 0 && (uintmax_t)devstat.st_gid != 0) {
			   
			   closedir(dirp);	
               return 0;
            }   
         }
         else {
      
            vdev_error("%s: lstat('%s'): '%s'\n", progname, str, strerror (errno));
            closedir(dirp);
            exit (EXIT_FAILURE);
         }
	  }
   }
   
   if(errno) {
      
      fprintf(stderr, "%s error reading current directory: %s\n", progname, strerror(errno));
      closedir(dirp);
      exit(EXIT_FAILURE);
   }
   
   closedir(dirp);

   while (1) {

      /* Initialize signals FD */
      if ((signal_fd = initialize_signals ()) < 0) {
      
         vdev_error ("%s: Couldn't initialize signals: '%s'\n", progname, strerror(errno));
         exit (EXIT_FAILURE);
      }

      /* Initialize inotify FD and the watch descriptors */
      if ((inotify_fd = initialize_inotify (argc, argv)) < 0) {
     
         vdev_error ("%s: Couldn't initialize inotify: '%s'\n", progname, strerror(errno));
         if (pfd[FD_POLL_SIGNAL].fd >= 0) {
          
            close (pfd[FD_POLL_SIGNAL].fd);
            pfd[FD_POLL_SIGNAL].fd = -1;
         }
         exit (EXIT_FAILURE);
      }  
    
      /* Setup polling */
      pfd[FD_POLL_SIGNAL].fd = signal_fd;
      pfd[FD_POLL_SIGNAL].events = POLLIN;
   
      pfd[FD_POLL_INOTIFY].fd = inotify_fd;
      pfd[FD_POLL_INOTIFY].events = POLLIN;
 
      // next event
      /* Block until there is something to be read */
      if (poll (pfd, FD_POLL_MAX, -1) < 0) {
      
         vdev_error ("%s: Couldn't poll(): '%s'\n", progname, strerror(errno));  
         shutdown_inotify (inotify_fd);
         close (signal_fd);
         exit (EXIT_FAILURE);
      }

      /* Signal received? */
      if (pfd[FD_POLL_SIGNAL].revents & POLLIN) {
         
         struct signalfd_siginfo fdsi;

         if (read(pfd[FD_POLL_SIGNAL].fd, &fdsi, sizeof (fdsi)) != sizeof (fdsi)) {
              
              vdev_error ("%s: Couldn't read signal, wrong size read: '%s'\n", progname, strerror (errno)); 
              shutdown_inotify (inotify_fd);
              close (signal_fd);
              exit (EXIT_FAILURE);
         }

         /* Break loop if we got the expected signal */
         if (fdsi.ssi_signo == SIGINT  ||
             fdsi.ssi_signo == SIGTERM ||
             fdsi.ssi_signo == SIGALRM) {

                printf("\n");
                break;
         }

         vdev_info ("%s: Received unexpected signal\n", progname);
      }   

      /* Inotify event received? */
      if (pfd[FD_POLL_INOTIFY].revents & POLLIN) {
         
         size_t length;
         char *aux;
         const size_t header_size = sizeof(int) + sizeof(uint32_t)*3;
         
         struct inot_event {
    
            int wd;
            uint32_t mask;
            uint32_t cookie;
            uint32_t len;
            char name[NAME_MAX+1];
         };

         char buf[sizeof(struct inot_event)];
         
         /* read inotify event(s) */
         length = read (pfd[FD_POLL_INOTIFY].fd, &buf, sizeof(struct inot_event));
         if(length >= header_size) { /* we've read at least one event */
     
            struct inot_event *inot_evt;
            
            /* We now have at least one inotify event in the buffer */
            for (aux=buf, inot_evt=(struct inot_event*)aux;
                 length>= header_size;
                 aux += (header_size + inot_evt->len), length -= (header_size + inot_evt->len), inot_evt=(struct inot_event*)aux) {
         
                    unsigned i=0;
                    
                    while (inot_evt->wd != monitors[i].wd)
                       i++;
            
                    if(inot_evt->mask & IN_ISDIR) {
               
                       char str[4097];            ///
                       struct stat devstat;
                       strncpy_t(str, sizeof(str), monitors[i].path, strlen(monitors[i].path));
                       strcat(str, "/");
                       strcat(str, inot_evt->name);
          
                       if (lstat(str, &devstat) == 0) {
              
                          //printf("created %s Ownership: UID=%ld   GID=%ld\n", inot_evt->name, (long) devstat.st_uid, (long) devstat.st_gid);
                          //if((uintmax_t)devstat.st_uid != 0 && (uintmax_t)devstat.st_gid != 0) {

                             is_userspace=true;
                             break;
                          //}
                       }
                       else {
      
                          vdev_error("%s: lstat('%s'): '%s'\n", progname, inot_evt->name, strerror (errno)); 
                          shutdown_inotify (inotify_fd);
                          close (signal_fd);
                          exit (EXIT_FAILURE);
                       }
                    }
            }  // for
         }
         else {
       
            vdev_error("%s: error reading inotify event: '%s'\n", progname, strerror (errno)); 
            shutdown_inotify (inotify_fd);
            close (signal_fd);
            exit (EXIT_FAILURE);
         }
    
         if(is_userspace) break;
     
      } 
   
   } // while
 
   shutdown_inotify (inotify_fd);
   close (signal_fd);
	
   return 0;
}



