
#include "libvdev/sglib.h"
#include "libvdev/util.h"
#include "libvdev/misc.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdbool.h>
#include <fcntl.h>
#include <mntent.h>
#include <sys/stat.h>

#include <blkid/blkid.h>

const char *progname = "vdev_modprobe";
const char *lockfile = "/run/vdev/vdev_modprobe.lock";

static int fd = -1;
static struct flock fl;

typedef char* cstr;

// prototypes
SGLIB_DEFINE_VECTOR_PROTOTYPES (cstr);
SGLIB_DEFINE_VECTOR_FUNCTIONS (cstr);

struct sysfs_scan_context {
   
   char* uevent_path;
   struct sglib_cstr_vector* frontier;
};

// 'buf' must be freed after usage
static
void get_sysfs_path (char **buf)
{
   struct mntent *e;
   FILE* fstab = setmntent("/etc/mtab", "r");
   if(!fstab) {
      
      vdev_error("%s: setmntent(): error trying to open /etc/mtab: '%s'\n", progname, strerror (errno));
      exit (EXIT_FAILURE);
   }  
 
   *buf = (char*)malloc(sizeof(char) * 32);
   if (!*buf) {

	  vdev_error ("%s: Memory allocation failure: '%s'\n", progname, strerror (errno));
	  exit (EXIT_FAILURE);
   }
   
   *buf[0] = '\0';
         
   while ((e = getmntent(fstab))) {
            
      if (!strcmp(e->mnt_type, "sysfs")) {

         sprintf( *buf, "%s", e->mnt_dir );
         break;
      }
   }
   
   endmntent(fstab);
}

// scan a directory in /sys/devices directory, to find its child directories, that are pushed back to scan_ctx->frontier
// return 0 on success
// return -ENOMEM on OOM
// return -errno on failure to stat
static
int scan_device_directory (char const* fp, void* cls)
{
   struct sysfs_scan_context *scan_ctx = (struct sysfs_scan_context*)cls;
   
   struct sglib_cstr_vector* frontier = scan_ctx->frontier;

   int rc = 0;  
   struct stat sb;
   char* fp_base = NULL;
   char* fp_dup = NULL;
   
   fp_base = rindex (fp, '/') + 1;
   
   if (fp_base == NULL)
      return 0;
   
   // skip . and .. 
   if (strcmp(fp_base, ".") == 0 || strcmp(fp_base, "..") == 0)
      return 0;
  
   // add directories
   rc = lstat (fp, &sb);
   if (rc != 0) {
      
      vdev_error ("%s: lstat('%s'): '%s'\n", progname, fp, strerror(errno));
      return -errno;
   }
   
   if (!S_ISDIR(sb.st_mode) && strcmp(fp_base, "uevent") != 0) {
      
      // not a directory, and not a uevent
      return 0;
   }
   
   fp_dup = vdev_strdup_or_null(fp);
   if (fp_dup == NULL)
      return -ENOMEM;
   
   if (S_ISDIR(sb.st_mode)) {
      
      rc = sglib_cstr_vector_push_back (frontier, fp_dup);
      if (rc != 0) {
      
         free (fp_dup);
         return rc;
      }
   }
   else {
      
      // this is a uevent; this directory is a device
      scan_ctx->uevent_path = fp_dup;
   }
   
   return 0;
}

// get a uevent from a uevent file 
// replace newlines with '\0', making the uevent look like it came from the netlink socket
// (i.e. so it can be parsed by vdev_linux_parse_request)
// return 0 on success
// return -ENOMEM on OOM
// return -errno on failure to stat or read
static
int vdev_read_uevent( char const* fp_uevent, char** ret_uevent_buf, size_t* ret_uevent_len )
{
   int rc = 0;
   struct stat sb;
   char* uevent_buf = NULL;
   size_t uevent_buf_len = 0;
   size_t uevent_len = 0;
   
   // get uevent size  
   rc = stat (fp_uevent, &sb);
   if (rc != 0) {
      
      vdev_error ("%s: stat('%s'): '%s'\n", progname, fp_uevent, strerror (errno));
      return -errno;
   }
   else
      uevent_buf_len = sb.st_size;
   
   // read the uevent as long as it's a regular file
   if (fp_uevent != NULL && (sb.st_mode & S_IFMT) == S_IFREG) {
      
      uevent_buf = VDEV_CALLOC (char, uevent_buf_len);
      if (uevent_buf == NULL)
         return -ENOMEM;
      
      rc = vdev_read_file( fp_uevent, uevent_buf, uevent_buf_len );
      if (rc != 0) {
         
         // failed in this 
         vdev_error ("%s: read_file('%s'): '%s'\n", progname, fp_uevent, strerror (errno));
         free (uevent_buf);
      }
      else {
         
         for (unsigned int i = 0; i < uevent_buf_len; i++) {
            
            if (uevent_buf[i] == '\n')               
               uevent_buf[i] = '\0';
         }
         
         // NOTE: the stat size is an upper-bound.  Find the exact number of bytes.
         for (uevent_len = 0; uevent_len < uevent_buf_len; ) {
            
            if (*(uevent_buf + uevent_len) == '\0')
               break;
            
            uevent_len += strlen( uevent_buf + uevent_len ) + 1;
         }
          
         *ret_uevent_buf = uevent_buf;
         *ret_uevent_len = uevent_len;
      }
   }
   
   return rc;
}

static
int modprobe (char const* fp_uevent)
{
   int rc = 0;
   struct stat sb;
   char* uevent_buf = NULL;
   size_t uevent_buf_len = 0;
   char* full_devpath = NULL;
   
   // get uevent
   rc = vdev_read_uevent( fp_uevent, &uevent_buf, &uevent_buf_len );
   if( rc != 0 ) {
      
      vdev_error ("%s: read_uevent('%s'): '%s'\n", progname, fp_uevent, strerror(errno));
      return rc;
   }
   
   if( uevent_buf_len == 0 ) {
      
      free (uevent_buf);
      return 0;
   }
   
   for (unsigned int i = 0; i < uevent_buf_len; ) {

      if (!strncmp(uevent_buf+i, "MODALIAS=", 9)) {
		 
		 FILE *pfin = NULL;
         pid_t pid;
         int wstatus;
         char cmd[1024]={0};
         char *tmp = rindex(uevent_buf + i, '=') + 1;
         
         strncpy_t (cmd, sizeof(cmd), "/sbin/modprobe -b -q -- \"", strlen("/sbin/modprobe -b -q -- \""));
         strcat(cmd, tmp);
         strcat(cmd, "\"");

         pfin = epopen(cmd, &pid);
         if(pfin) { 
            fclose(pfin);
            waitpid(pid, &wstatus, 0);
         }

         //if (wstatus != 0)
            //fprintf(stderr, "Modprobe failed: %s\n", tmp);
      }
           
      i += strlen(uevent_buf + i) + 1;
   }
   
   // Try to create a lock file. Bear in mind that the system still may be read only at this point.
   // If so, we ignore the failed atempt trying again when loading the next driver 
   if (fd == -1) {
      
      fd = open(lockfile, O_RDWR|O_CREAT, 0640); 
      if (fd < 0) {

	     switch(errno) {
	        
	        case ENOMEM:
	           return -ENOMEM;
	        
	        default:
               return rc;
	     }      
      }
   }
   else   
      return rc;   

   memset(&fl, 0, sizeof(fl));
   
   fl.l_type = F_WRLCK; // Set a write lock
   fl.l_whence = SEEK_SET; // Start at beginning of file
   fl.l_start = 0; // Offset from beginning of file
   fl.l_len = 0; // Lock entire file
   
   // F_SETLK(W) ignores it; F_OFD_SETLK(W) requires it to be zero
   fl.l_pid = 0; //getpid(); // Set process ID
   
   // F_SETLKW specifies blocking mode
   fcntl(fd, F_SETLKW, &fl); // Set the lock
   
   return rc;
}

/**
 * \brief Variadic function
 */
static
int find_devices_at_frontier (const char *sysfs_path, const char *device_frontier, struct sglib_cstr_vector *uevent_paths, struct sglib_cstr_vector *subsystems)
{
   int rc = 0;
   struct sglib_cstr_vector frontier;
   struct sysfs_scan_context scan_ctx;

   sglib_cstr_vector_init (&frontier);
   
   memset (&scan_ctx, 0, sizeof(struct sysfs_scan_context));
   
   scan_ctx.frontier = &frontier;

   rc = vdev_load_all (device_frontier, scan_device_directory, &scan_ctx);
   if (rc != 0) {
      
      vdev_error ("%s: vdev_load_all('%s'): '%s'\n", progname, device_frontier, strerror(errno));
      sglib_cstr_vector_free (&frontier);      
      return rc;
   }
   
   while (1) {

      
      unsigned long len = sglib_cstr_vector_size (&frontier);
      if (len == 0) break;
   
      char* dir = NULL;
      
      dir = sglib_cstr_vector_at (&frontier, len - 1);
      sglib_cstr_vector_set (&frontier, NULL, len - 1);
      
      sglib_cstr_vector_pop_back (&frontier);
      
      // scan for more devices 
      rc = vdev_load_all (dir, scan_device_directory, &scan_ctx);
      if (rc != 0) {
         
         vdev_error ("%s: vdev_load_all('%s'): '%s'\n", progname, dir, strerror(errno));
         free (dir);
         break;
      }
      
      // is one of them a uevent?
      if (scan_ctx.uevent_path) {
		  
		 const char *str;
		 bool is_ok = false;
         
         // yup--remember it
         char* uevent_path = vdev_strdup_or_null (scan_ctx.uevent_path);
         
         if (!uevent_path) {
            
            free (dir);
            free (scan_ctx.uevent_path);
            scan_ctx.uevent_path = NULL;
            
            rc = -ENOMEM;
            break;
         }
      
         for( unsigned i = 0; i < sglib_cstr_vector_size(subsystems); i++ ) {
			 
            if (sysfs_get_parent_with_subsystem (uevent_path, sglib_cstr_vector_at(subsystems, i)) && !sysfs_get_parent_with_subsystem (uevent_path, "input")) {
               
               is_ok = true;
               break;
            }
         }

         if (!is_ok) {

            int r = 0;			
            char s[1024]={0};
            struct stat sb;
            const char *dname = dirname (vdev_strdup_or_null(uevent_path));
            strncpy_t(s, sizeof(s), dname, strlen(dname));
            strcat(s, "/boot_vga");
            r = stat( s, &sb );
            if( r != 0 )
              continue;
		 }

         rc = sglib_cstr_vector_push_back (uevent_paths, uevent_path);
         if (rc < 0) {
            
            free (uevent_path);
            free (dir);
            free (scan_ctx.uevent_path);
            scan_ctx.uevent_path = NULL;
            
            break;
         }
         
         free (scan_ctx.uevent_path);
         scan_ctx.uevent_path = NULL;
         
         rc = modprobe (uevent_path);
         if (rc != 0)
            vdev_error ("%s: modprobe('%s'): '%s'\n", progname, uevent_path, strerror(errno));
      }
      
      free (dir);
   }
   
   sglib_cstr_vector_free (&frontier);

   return rc;
}

static
int find_devices (char *sysfs_mountpoint, struct sglib_cstr_vector* uevent_paths, int count, ...)
{
   int rc = 0;
   struct sglib_cstr_vector frontier;
   struct sglib_cstr_vector subsystems;
   struct sysfs_scan_context scan_ctx;  
   va_list items;
   
   memset (&scan_ctx, 0, sizeof(struct sysfs_scan_context));
      
   char* devroot = NULL;

   sglib_cstr_vector_init (&frontier);
   sglib_cstr_vector_init (&subsystems);
   
   scan_ctx.frontier = &frontier;
   
   devroot = vdev_fullpath (sysfs_mountpoint, "/devices", NULL);     // devroot <= /sys/devices
   if( devroot == NULL )      
      return -ENOMEM;
    
   rc = vdev_load_all (devroot, scan_device_directory, &scan_ctx);
   if( rc != 0 ) {
      
      vdev_error ("%s: vdev_load_all('%s'): '%s'\n", progname, devroot, strerror(errno)); 
      free (devroot);    
      sglib_cstr_vector_free (&frontier);      
      return rc;
   }
   
   free (scan_ctx.uevent_path);
   scan_ctx.uevent_path = NULL;
	
   va_start(items, count);  /* Requires the last fixed parameter (to get the address) */
         
   while (count--) {

      rc = sglib_cstr_vector_push_back (&subsystems, va_arg(items, char*));
      if (rc < 0) {
            
         sglib_cstr_vector_free (&frontier);   
         sglib_cstr_vector_free (&subsystems);
         return rc;
      }
   }
         
   va_end(items);
 
   for (unsigned long i = 0; i < sglib_cstr_vector_size (&frontier); i++)
      find_devices_at_frontier (sysfs_mountpoint, sglib_cstr_vector_at (&frontier, i), uevent_paths, &subsystems);
   
   sglib_cstr_vector_free (&frontier);   
   sglib_cstr_vector_free (&subsystems);
   
   return rc;
}


/**
 * \brief This function will daemonize this app
 */
static
void daemonize()
{
   int ret;
   pid_t pid = 0;
   int fd;

   /* Fork off the parent process.
    * The first fork will change our pid
    * but the sid and pgid will be the
    * calling process
    */
   pid = fork();

   /* An error occurred */
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }

   /* Fork off for the second time.
    * The magical double fork. We're the session
    * leader from the code above. Since only the
    * session leader can take control of a tty
    * we will fork and exit the session leader.
    * Once the fork is done below and we use
    * the child process we will ensure we're
    * not the session leader, thus, we cannot
    * take control of a tty.
    */
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }

   /* On success: The child process becomes session leader */
   if (setsid() < 0) {
      exit(EXIT_FAILURE);
   }

   /* Ignore signal sent from child to parent process */
   signal(SIGCHLD, SIG_IGN);

   /* Fork off for the second time*/
   pid = fork();

   /* An error occurred */
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }

   /* Success: Let the parent terminate */
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }
   
   /* Set new file permissions */
   umask(0);

   /* Change the working directory to /tmp */
   /* or another appropriated directory */
   ret = chdir("/tmp");
   if(ret != 0) exit(EXIT_FAILURE);

   /* Reopen stdin (fd = 0), stdout (fd = 1), stderr (fd = 2) */
   stdin  = fopen("/dev/null", "r");
   stdout = fopen("/dev/null", "w+");
   stderr = fopen("/dev/null", "w+");
}

int main (int argc, char **argv)
{
   int rc, flag;
   struct sglib_cstr_vector uevent_paths;
   char sysfs_mountpoint[PATH_MAX+1];
   char * buf = NULL; 
   
   get_sysfs_path (&buf);
   if(buf && buf[0] != '\0') {
 
      snprintf (sysfs_mountpoint, PATH_MAX, buf);       
      free (buf);   
   }
   else {
	  
	  vdev_error ("%s: Cannot get sysfs path\n", progname);
      exit (EXIT_FAILURE);
   }
  
   daemonize();
   
   sglib_cstr_vector_init (&uevent_paths);
   rc = find_devices (sysfs_mountpoint, &uevent_paths, 3, "drm", "pci", "drivers" );
   sglib_cstr_vector_free (&uevent_paths);

   if (fd > 0) {
   
      fl.l_type = F_UNLCK; // Release the lock
      fcntl (fd, F_SETLK, &fl); // Unlock the file
      
      close (fd);
      unlink (lockfile);
   }
   	
   return 0;
}



