
#include "libvdev/sglib.h"
#include "libvdev/util.h"
#include "libvdev/misc.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <errno.h>
#include <unistd.h>
#include <mntent.h>
#include <fstab.h>
#include <poll.h>
#include <limits.h>
#include <signal.h>
#include <sys/signalfd.h>
#include <sys/inotify.h>
#include <sys/stat.h>

#include <blkid/blkid.h>
#include <libudev.h>

const char *progname = "vdev_reset_usb";

/* Structure to keep track of monitored directories */
typedef struct {
  
   /* Path of the directory */
   char *path;
  
   /* inotify watch descriptor */
   int wd;
  
} monitored_t;

/* Size of buffer to use when reading inotify events */
#define INOTIFY_BUFFER_SIZE 8192

/* Enumerate list of FDs to poll */
enum {
  
   FD_POLL_SIGNAL = 0,
   FD_POLL_INOTIFY,
   FD_POLL_UDEV,
   FD_POLL_MAX
};

/* Array of directories being monitored */
static monitored_t *monitors;
static int n_monitors;

static
void shutdown_inotify (int inotify_fd)
{
   for (unsigned i = 0; i < n_monitors; ++i) {
      
      free (monitors[i].path);
      inotify_rm_watch (inotify_fd, monitors[i].wd);
   }
   
   free (monitors);
}

static
int initialize_inotify (int argc, char **argv)
{
   int inotify_fd;

   /* Create new inotify device */
   if ((inotify_fd = inotify_init ()) < 0) {
      
      vdev_error("%s: Couldn't setup new inotify device: '%s'\n", progname, strerror (errno));
      return -1;
   }

   /* Allocate array of monitor setups */
   n_monitors = argc - 1;
   monitors = malloc (n_monitors * sizeof (monitored_t));

   /* Loop all input directories, setting up watches */
   for (unsigned i = 0; i < n_monitors; ++i) {
      
      monitors[i].path = vdev_strdup_or_null (argv[i + 1]);
      if ((monitors[i].wd = inotify_add_watch (inotify_fd, monitors[i].path, IN_CREATE)) < 0) {
        
          vdev_error ("%s: Couldn't add monitor in directory '%s': '%s'\n", progname, monitors[i].path, strerror (errno));
          exit (EXIT_FAILURE);
      }
      
      //printf ("Started monitoring directory '%s'...\n", monitors[i].path);
   }

   return inotify_fd;
}

static
int initialize_signals (void)
{
   int signal_fd;
   sigset_t sigmask;

   /* We want to handle SIGINT and SIGTERM in the signal_fd, so we block them. */
   sigemptyset (&sigmask);
   sigaddset (&sigmask, SIGINT);
   sigaddset (&sigmask, SIGTERM);

   if (sigprocmask (SIG_BLOCK, &sigmask, NULL) < 0) {
      
      vdev_error ("%s: Couldn't block signals: '%s'\n", progname, strerror (errno));
      return -1;
   }

   /* Get new FD to read signals from it */
   if ((signal_fd = signalfd (-1, &sigmask, 0)) < 0) {
      
      vdev_error ("%s: Couldn't setup signal FD: '%s'\n", progname, strerror (errno));
      return -1;
   }

   return signal_fd;
}

#define BLOCK_SIZE 512

typedef char* cstr;

// prototypes
SGLIB_DEFINE_VECTOR_PROTOTYPES (cstr);
SGLIB_DEFINE_VECTOR_FUNCTIONS (cstr);



static
bool parse_device (const char *devnode)
{

   struct fstab *fstab_ent;
   bool res = true;

   if (setfsent() != 1) /* Try to open fstab */ {
      
      return false;
      
   } else {      

      DIR *dirp;
      struct dirent *E;
      const char *tmp;
      blkid_probe pr;
      blkid_partlist ls; 
      int nparts;
      char sysblockdir [1024];
      
      pr = blkid_new_probe_from_filename(devnode);
      if (!pr) {

         printf ("%s: Failed to open %s: '%s'\n", progname, devnode, strerror (errno));
      }

      ls = blkid_probe_get_partitions(pr);
      if(!ls) {
		 
		 blkid_free_probe(pr); 
         return true;
      }
         
      nparts = blkid_partlist_numof_partitions(ls);
      if (nparts <= 0) {
		 
		 blkid_free_probe(pr); 
         return true;
      }   
	   
	  tmp = rindex (devnode, '/') + 1;
      strncpy_t (sysblockdir, sizeof(sysblockdir), "/sys/block/", strlen("/sys/block/"));
      strcat(sysblockdir, tmp);
 
      dirp = opendir(sysblockdir);
      if(!dirp) {
      
         printf("%s error opening current directory: %s\n", progname, strerror(errno));
         blkid_free_probe(pr);
         exit(EXIT_FAILURE);
      }

      while( (errno=0, E=readdir(dirp)) ) {
	  
         char partpth[1024];
         strncpy_t (partpth, sizeof(partpth), sysblockdir, strlen(sysblockdir));
         strcat(partpth, "/");
         strcat(partpth, E->d_name);
         strcat(partpth, "/partition");
      
         if(!strncmp(E->d_name, tmp, strlen(tmp)) && (access(partpth, F_OK) == 0)) {

            const char *uuid;
            char node[1024];
            
            strncpy_t (node, sizeof(node), "/dev/", strlen("/dev/"));
            strcat(node, E->d_name);
            
            pr = blkid_new_probe_from_filename(node);
            blkid_do_probe(pr);

            blkid_probe_lookup_value(pr, "UUID", &uuid, NULL);
      
            while ((fstab_ent = getfsent()) != NULL) {
            
               char *aux = rindex (fstab_ent->fs_spec, '=') + 1;
               if (!strcmp(uuid, aux)) {

                  endfsent(); /* Close fstab */
                  blkid_free_probe(pr);
                  closedir(dirp);
                  return false;
               }
            }
         
            endfsent(); /* Close fstab */
         }
	  }
	  
	  closedir(dirp);

      blkid_free_probe(pr);
   }
   
   return res;
}

// buf must be freed after usage.
static
void get_bus_usb_path ( char const* syspath, char **buf )
{
   char path[4097];
   char subsystem_path[4097];
   char subsystem_link[4097];
   char* parent_device = NULL;
   size_t parent_device_len = 0;
   char *p[] = { "busnum", "devnum", "authorized", "uevent", (char*)0 };
    
   *buf = (char*)malloc(sizeof(char) * 4097);
   if (!*buf) {
      
      vdev_error ("%s: Memory allocation failure: '%s'\n", progname, strerror (errno));
      exit (EXIT_FAILURE);
   }
   
   *buf[0] = '\0';
   
   memset( subsystem_path, 0, 4097 );
   memset( subsystem_link, 0, 4097 );
      
   memset( path, 0, 4097 );
   strncpy_t( path, sizeof(path), syspath, strlen(syspath) );
 
   while(1) {  
      
      struct stat sb;
      bool res = true;

      for( unsigned long i = 0; i < sizeof(p)/sizeof(char*) - 1; i++ ) {

         int r;
         sprintf( *buf, "%s/%s", path, p[i] );
         r = stat( *buf, &sb );
         if( r != 0 ) {
         
            res = false;
         }
     }
      
      if(!res) {
        
         // get parent device 
         int r = sysfs_get_parent_device( path, &parent_device, &parent_device_len );
         if( r != 0 )
            break;
         
         memset( path, 0, 4097 );
         strncpy_t (path, sizeof(path), parent_device, parent_device_len);
      }
      else
         break;
   }

   return;
}

static
int authorize_device (const char *path, int flag)
{ 
   int rc;
   char cmd[4097];
   FILE *pfin = NULL;
   pid_t pid;
      
   memset( cmd, 0, 4097 );
   strncpy_t(cmd, sizeof(cmd), "echo ", strlen("echo "));
   
   if(flag == 0)
      strcat(cmd, "0 > ");
   else
      strcat(cmd, "1 > ");
   
   strcat(cmd, dirname(vdev_strdup_or_null(path)));   
   strcat(cmd, "/authorized");
 
   pfin = epopen(cmd, &pid);
   if(pfin) { 
      
      fclose(pfin);
      waitpid(pid, &rc, 0);
   }

   return rc;
}

/**
 * \brief This function will daemonize this app
 */
static
void daemonize()
{
   int ret;
   pid_t pid = 0;
   int fd;

   /* Fork off the parent process.
    * The first fork will change our pid
    * but the sid and pgid will be the
    * calling process
    */
   pid = fork();

   /* An error occurred */
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }

   /* Fork off for the second time.
    * The magical double fork. We're the session
    * leader from the code above. Since only the
    * session leader can take control of a tty
    * we will fork and exit the session leader.
    * Once the fork is done below and we use
    * the child process we will ensure we're
    * not the session leader, thus, we cannot
    * take control of a tty.
    */
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }

   /* On success: The child process becomes session leader */
   if (setsid() < 0) {
      exit(EXIT_FAILURE);
   }

   /* Ignore signal sent from child to parent process */
   signal(SIGCHLD, SIG_IGN);

   /* Fork off for the second time*/
   pid = fork();

   /* An error occurred */
   if (pid < 0) {
      exit(EXIT_FAILURE);
   }

   /* Success: Let the parent terminate */
   if (pid > 0) {
      exit(EXIT_SUCCESS);
   }
   
   /* Set new file permissions */
   umask(0);

   /* Change the working directory to /tmp */
   /* or another appropriated directory */
   ret = chdir("/tmp");
   if(ret != 0) exit(EXIT_FAILURE);

   /* Reopen stdin (fd = 0), stdout (fd = 1), stderr (fd = 2) */
   stdin  = fopen("/dev/null", "r");
   stdout = fopen("/dev/null", "w+");
   stderr = fopen("/dev/null", "w+");
}

int main(int argc, char **argv)
{
   struct udev *udev = NULL;
   struct udev_device *dev = NULL;
   struct udev_enumerate *enumerate = NULL;
   struct udev_list_entry *devices, *dev_list_entry;
   
   char pathbuf[PATH_MAX+1];
   struct udev_monitor* monitor = NULL;
   struct pollfd pfd[FD_POLL_MAX];   
   struct sglib_cstr_vector items;
  
   bool is_userspace=false;
 
   int signal_fd, inotify_fd;
   int monitor_fd = 0;
   
   /* Input arguments... */
   if (argc < 2) {
      
      fprintf (stderr, "Usage: %s  directory1 [directory2 ...]\n", argv[0]);
      exit (EXIT_FAILURE);
   }
   
   daemonize();
  
   sglib_cstr_vector_init (&items);

   /* create udev object */
   udev = udev_new ();
   if (!udev) {

      vdev_error("%s: Cannot create udev context: '%s'\n", progname, strerror (errno));
      return 1;
   }

   /* create enumerate object */
   enumerate = udev_enumerate_new (udev);
   if (!enumerate) {

      vdev_error ("%s: Cannot create enumerate context: '%s'\n", progname, strerror (errno));
      return 1;
   }

   udev_enumerate_add_match_subsystem (enumerate, "block");
   udev_enumerate_scan_devices (enumerate);

   /* fillup device list */
   devices = udev_enumerate_get_list_entry (enumerate);
   if (!devices) {
 
      vdev_error ("%s: Failed to get device list: '%s'\n", progname, strerror (errno));
      udev_enumerate_unref (enumerate);
      udev_unref (udev);   
      return 1;
   }

   udev_list_entry_foreach (dev_list_entry, devices) {

      const char *path, *tmp;
      unsigned long long disk_size = 0;
      unsigned short int block_size = BLOCK_SIZE;
      struct udev_device *dev = NULL;

      path = udev_list_entry_get_name (dev_list_entry);
      dev = udev_device_new_from_syspath (udev, path);
      if (dev == NULL)
          continue;

      /* skip if device/disk is a partition or loop device */
      if (strcmp(udev_device_get_subsystem( dev ), "block") == 0 &&
          strncmp(udev_device_get_devtype (dev), "partition", 9) &&
          strncmp(udev_device_get_sysname (dev), "loop", 4)) {
      
          /* skip if some of its partions appears in /etc/fstab */
          bool parse_ok = false;
          const char *devnode = udev_device_get_devnode(dev);

          parse_ok = parse_device (devnode);
          if (parse_ok) {
              
             char *buf = NULL;
             get_bus_usb_path (udev_device_get_syspath( dev ), &buf);
             if (buf && buf[0] != '\0') {
                     
                authorize_device (buf, 0);
                sglib_cstr_vector_push_back (&items, vdev_strdup_or_null (buf));        
                free(buf);   
             }
             else if (buf)
                free (buf);
          }   
      }

      udev_device_unref (dev);
   }

   udev_enumerate_unref (enumerate);
  
   /* NETLINK */
    
   udev_monitor_fs_events_path( "", pathbuf, 0 );
   
   monitor = udev_monitor_new_from_netlink ( udev, "udev" );
   if( monitor == NULL ) {
 
      // OOM or error
      vdev_error ("%s: Couldn't initialize udev monitor: '%s'\n", progname, strerror(errno));
      sglib_cstr_vector_free (&items);
      udev_unref (udev);
      exit(EXIT_FAILURE);
   }
   
   monitor_fd = udev_monitor_get_fd (monitor);
   if( monitor_fd < 0 ) {
 
      vdev_error ("%s: udev_monitor_get_fd: '%s'\n", progname, strerror(errno));
      sglib_cstr_vector_free (&items);
      udev_unref (udev);
      exit (EXIT_FAILURE);
   }

   /* Initialize signals FD */
   if ((signal_fd = initialize_signals ()) < 0) {
     
      vdev_error ("%s: Couldn't initialize signals: '%s'\n", progname, strerror(errno));
      sglib_cstr_vector_free (&items);
      udev_monitor_fs_destroy (monitor);
      udev_unref (udev);
      exit (EXIT_FAILURE);
   }

   /* Initialize inotify FD and the watch descriptors */
   if ((inotify_fd = initialize_inotify (argc, argv)) < 0) {
 
      vdev_error ("%s: Couldn't initialize inotify: '%s'\n", progname, strerror(errno));
      if( pfd[FD_POLL_SIGNAL].fd >= 0 ) {
          
            close( pfd[FD_POLL_SIGNAL].fd );
            pfd[FD_POLL_SIGNAL].fd = -1;
      }
      
      sglib_cstr_vector_free (&items);
      udev_monitor_fs_destroy (monitor);
      udev_unref (udev);
      exit (EXIT_FAILURE);
   }  
    
   /* Setup polling */
   pfd[FD_POLL_SIGNAL].fd = signal_fd;
   pfd[FD_POLL_SIGNAL].events = POLLIN;
   
   pfd[FD_POLL_INOTIFY].fd = inotify_fd;
   pfd[FD_POLL_INOTIFY].events = POLLIN;
   
   pfd[FD_POLL_UDEV].fd = monitor_fd;
   pfd[FD_POLL_UDEV].events = POLLIN;
   
   daemonize();

   while(1) {
 
      // next event
      /* Block until there is something to be read */
      if (poll (pfd, FD_POLL_MAX, -1) < 0) {

          vdev_error ("%s: Couldn't poll(): '%s'\n", progname, strerror (errno));
          goto Exit;
      }

      /* Signal received? */
      if (pfd[FD_POLL_SIGNAL].revents & POLLIN) {
         
         struct signalfd_siginfo fdsi;

         if (read (pfd[FD_POLL_SIGNAL].fd, &fdsi, sizeof (fdsi)) != sizeof (fdsi)) {
 
              vdev_error ("%s: Couldn't read signal, wrong size read: '%s'\n", progname, strerror (errno));
              goto Exit;
         }

         /* Break loop if we got the expected signal */
         if (fdsi.ssi_signo == SIGINT  ||
             fdsi.ssi_signo == SIGTERM ||
             fdsi.ssi_signo == SIGALRM) {

                printf("\n");
                break;
         }

         vdev_info ("%s: Received unexpected signal\n", progname);
      }   

      /* Inotify event received? */
      if (pfd[FD_POLL_INOTIFY].revents & POLLIN) {
         
         size_t length;
         char *aux;
         const size_t header_size = sizeof(int) + sizeof(uint32_t)*3;
         
         struct inot_event {
    
            int wd;
            uint32_t mask;
            uint32_t cookie;
            uint32_t len;
            char name[NAME_MAX+1];
         };

         char buf[sizeof(struct inot_event)];
         
         /* read inotify event(s) */
         length = read(pfd[FD_POLL_INOTIFY].fd, &buf, sizeof(struct inot_event));
         if(length >= header_size) { /* we've read at least one event */
     
            struct inot_event *inot_evt;
            
            /* We now have at least one inotify event in the buffer */
            for (aux=buf, inot_evt=(struct inot_event*)aux;
                 length>= header_size;
                 aux += (header_size + inot_evt->len), length -= (header_size + inot_evt->len), inot_evt=(struct inot_event*)aux) {
         
                    unsigned i=0;
                    
                    while (inot_evt->wd != monitors[i].wd)
                       i++;
            
                    if(inot_evt->mask & IN_ISDIR) {
               
                       char str[4097];            ///
                       struct stat devstat;
                       strncpy_t(str, sizeof(str), monitors[i].path, strlen(monitors[i].path));
                       strcat(str, "/");
                       strcat(str, inot_evt->name);
          
                       if (lstat(str, &devstat) == 0) {
              
                          //printf("created %s Ownership: UID=%ld   GID=%ld\n", inot_evt->name, (long) devstat.st_uid, (long) devstat.st_gid);
                          if((uintmax_t)devstat.st_uid != 0 && (uintmax_t)devstat.st_gid != 0) {
                             is_userspace=true;                  
                             break;
                          }   
                       }
                       else {
      
                          vdev_error("%s: lstat('%s'): '%s'\n", progname, inot_evt->name, strerror (errno));
                          goto Exit;
                       }
                    }
            }  // for
         }
         else {
       
            vdev_error("%s: error reading inotify event: '%s'\n", progname, strerror (errno));
            goto Exit;
         }
    
         if(is_userspace) break;
     
      }
      
      // get devices 
      while(1) {
         
         dev = udev_monitor_receive_device (monitor);
         if (dev == NULL)
            break;

         /* skip if device/disk is a partition or loop device */
         if (strcmp(udev_device_get_action( dev ), "add") == 0 &&
             strcmp(udev_device_get_subsystem( dev ), "block") == 0 &&
             strncmp(udev_device_get_sysname(dev), "loop", 4) &&
             strncmp(udev_device_get_devtype(dev), "partition", 9)) {
              
             char *buf = NULL;
             get_bus_usb_path (udev_device_get_syspath( dev ), &buf);
             if(buf && buf[0] != '\0') {

                authorize_device (buf, 0);
                sglib_cstr_vector_push_back (&items, vdev_strdup_or_null (buf));        
                free(buf);   
             }
         }
        
         udev_device_unref( dev );   
      }
   }
   
   sleep (5);
   
   for (unsigned long i = 0; i < sglib_cstr_vector_size (&items); i++) {

      authorize_device (sglib_cstr_vector_at (&items, i), 1);
   }
 
Exit:
  
   shutdown_inotify (inotify_fd);
      
   // we do not close pfd[FD_POLL_UDEV] because it's closed afterwards by udev_monitor_fs_destroy()
   // Hence, we exit the loop when i<FP_POLL_MAX - 1
   for (unsigned i = 0; i < FD_POLL_MAX - 1; i++) {

      if( pfd[i].fd >= 0 ) {
          
         close( pfd[i].fd );
         pfd[i].fd = -1;
      }
   }
   
   sglib_cstr_vector_free (&items);
   udev_monitor_fs_destroy (monitor);

   udev_unref (udev);

   return 0;
}
